(* Semantic checking for the E-CATZ compiler *)

open Ast
open Sast

module StringMap = Map.Make(String)

(* Semantic checking of the AST. Returns an SAST if successful,
   throws an exception if something is wrong.

   Check each global variable, then check each function *)

let check (statements, functions) =

  (* Verify a list of bindings has no void types or duplicate names *)
  let check_binds (kind : string) (binds : bind list) =
    List.iter (function
	(Void, b) -> raise (Failure ("illegal void " ^ kind ^ " " ^ b))
      | _ -> ()) binds;
    let rec dups = function
        [] -> ()
      |	((_,n1) :: (_,n2) :: _) when n1 = n2 ->
	  raise (Failure ("duplicate " ^ kind ^ " " ^ n1))
      | _ :: t -> dups t
    in dups (List.sort (fun (_,a) (_,b) -> compare a b) binds)
  in

  (**** Check functions ****)

  (* Collect function declarations for built-in functions with one argument: no bodies *)
  let built_in_decls_partial_partial =
    let add_bind map (name, ty) = StringMap.add name {
      typ = Void;
      fname = name;
      formals = [(ty, "x")];
      body = [] } map
    in List.fold_left add_bind StringMap.empty [ ("print", Int); ("prints", String);
                                                ("printb", Bool); ("printf", Float); ]
  in

  (* Built-ins with 2 arguments *)
  let built_in_decls_partial =
    let add_bind map (name, ty1, ty2, r) = StringMap.add name {
      typ = r;
      fname = name;
      formals = [(ty1, "x"); (ty2, "y")];
      body = [] } map
    in List.fold_left add_bind built_in_decls_partial_partial
      [ ("randInt", Int, Int, Int); ("write", Array Note, String, Void);
         ("writeN", Array (Array Note), String, Void); ]
  in

  (* Built-ins with no arguments *)
  let built_in_decls =
   StringMap.add "seed" {
    typ = Void;
    fname = "seed";
    formals = [];
    body = [] } built_in_decls_partial
    in

  (* Add function name to symbol table *)
  let add_func map fd =
    let built_in_err = "function " ^ fd.fname ^ " may not be defined"
    and dup_err = "duplicate function " ^ fd.fname
    and make_err er = raise (Failure er)
    and n = fd.fname (* Name of the function *)
    in match fd with (* No duplicate functions or redefinitions of built-ins *)
         _ when StringMap.mem n built_in_decls -> make_err built_in_err
       | _ when StringMap.mem n map -> make_err dup_err
       | _ ->  StringMap.add n fd map
  in

  (* Collect all function names into one symbol table *)
  let function_decls = List.fold_left add_func built_in_decls functions
  in

  (* Return a function from our symbol table *)
  let find_func s =
    try StringMap.find s function_decls
    with Not_found -> raise (Failure ("unrecognized function " ^ s))
  in

  (* Raise an exception if the given rvalue type cannot be assigned to
     the given lvalue type *)
  let check_assign lvaluet rvaluet err =
     if (lvaluet = rvaluet || (lvaluet = Note && rvaluet = Pitch)) then lvaluet else raise (Failure err)
  in

  (* Return a variable from our local symbol table *)
  let type_of_identifier s symbols =
    try StringMap.find s symbols
    with Not_found -> raise (Failure ("undeclared identifier " ^ s))
  in
  (* Return a semantically-checked expression, i.e., with a type *)
  let rec expr symbols = function
      PLiteral l -> (Pitch, SPLiteral l)
    | StrLiteral l -> (String, SStrLiteral l)
    | RLiteral l -> (Rhythm, SRLiteral l)
    | Literal  l -> (Int, SLiteral l)
    | Fliteral l -> (Float, SFliteral l)
    | BoolLit l  -> (Bool, SBoolLit l)
    | Noexpr     -> (Void, SNoexpr)
    | Id s       -> (type_of_identifier s symbols, SId s)
    | Assign(var, e) as ex ->
              let (lt, var') = expr symbols var
              and (rt, e') = expr symbols e in
              let err = "illegal assignment " ^ string_of_typ lt ^ " = " ^
                string_of_typ rt ^ " in " ^ string_of_expr ex
              in (check_assign lt rt err, SAssign((lt, var'), (rt, e')))
    | Unop(op, e) as ex -> 
          let (t, e') = expr symbols e in
          let ty = match op with
            Neg when t = Int || t = Float -> t
          | Not when t = Bool -> Bool
          | _ -> raise (Failure ("illegal unary operator " ^ 
                                 string_of_uop op ^ string_of_typ t ^
                                 " in " ^ string_of_expr ex))
          in (ty, SUnop(op, (t, e')))
    | Binop(e1, op, e2) as e ->
              let (t1, e1') = expr symbols e1
              and (t2, e2') = expr symbols e2 in
              (* All binary operators require operands of the same type *)
              let same = t1 = t2 in
              (* Determine expression type based on operator and operand types *)
              let ty = match op with
                Add | Sub | Mult | Div | Mod when same && t1 = Int   -> Int
              | Add | Sub | Mult | Div when same && t1 = Float -> Float
              | Equal | Neq  when same && (t1 = Int || t1 = Float || t1 = Pitch || t1 = Rhythm) -> Bool
              | Less | Leq | Greater | Geq
                         when same && (t1 = Int || t1 = Float) -> Bool
              | And | Or when same && t1 = Bool -> Bool
              | _ -> raise (
            Failure ("illegal binary operator " ^
                           string_of_typ t1 ^ " " ^ string_of_op op ^ " " ^
                           string_of_typ t2 ^ " in " ^ string_of_expr e))
              in (ty, SBinop((t1, e1'), op, (t2, e2')))
    | Call(fname, args) as call ->
        let fd = find_func fname in
        let param_length = List.length fd.formals in
        if List.length args != param_length then
          raise (Failure ("expecting " ^ string_of_int param_length ^
                          " arguments in " ^ string_of_expr call))
        else let check_call (ft, _) e =
          let (et, e') = expr symbols e in
          let err = "illegal argument found " ^ string_of_typ et ^
            " expected " ^ string_of_typ ft ^ " in " ^ string_of_expr e
          in (check_assign ft et err, e')
        in
        let args' = List.map2 check_call fd.formals args
        in (fd.typ, SCall(fname, args'))
    (* Notes must be created with exactly two arguments, pitch and rhythm *)
    | CreateNote(args) as call ->
        if List.length args != 2 then
          raise (Failure ("expecting " ^ string_of_int 2 ^
                          " arguments in " ^ string_of_expr call))
        else let check_call ft e =
          let (et, e') = expr symbols e in
          let err = "illegal argument found " ^ string_of_typ et ^
            " expected " ^ string_of_typ ft ^ " in " ^ string_of_expr e
          in (check_assign ft et err, e')
        in
        let args' = List.map2 check_call [Pitch; Rhythm] args
        in (Note, SCreateNote(args'))
    | NoteAccess(s, p) ->
        let (typ, s') = expr symbols s in
        (* Make sure that dot access is being used on a note *)
        if (typ=Note) then
          let rtyp = match p with
              P -> Pitch
            | R -> Rhythm
          in
          (rtyp, SNoteAccess((typ, s'), p))
        else
          raise (Failure ("Accessing pitch or rhythm may only be used on identifier of type Note. " ^ string_of_expr s ^ " is not a Note." ))
    | NoteAssign(s, p, e) as ex ->
        let (typ, s') = expr symbols s in
        (* Make sure that dot access is being used on a note *)
        if (typ=Note) then
          let lt = match p with
              P -> Pitch
            | R -> Rhythm
          in
          let (rt, e') = expr symbols e in
          let err = "illegal assignment " ^ string_of_typ lt ^ " = " ^
            string_of_typ rt ^ " in " ^ string_of_expr ex
          in (check_assign lt rt err, SNoteAssign((typ, s'), p, (rt, e')))
        else
          raise (Failure ("Attempting to assign a Note to an identifier of the wrong type" ))
    | ArrayLength(s) ->
       let (typ, s') = expr symbols s in
       (* Check that we're trying to get the length of an array*)
       let _ = match typ with
          Array t -> t
          | _ ->  raise (Failure ("Only arrays have lengths you fool. " ^ string_of_expr s ^ " is not an array." ))
       in (Int, SArrayLength((typ, s')))
    | ArrayLiteral(l) ->
        let param_length = List.length l in
        (* if someone types in [], it raises an error *)
        if (param_length = 0) then
          raise (Failure ("Array literals must have at least one argument"))
        else
          (* Assume the first type of the array is the right type *)
          let (first_type, _) = expr symbols (List.hd l) in
          (* If a pitch literal appears in an array literal, treat it as a note *)
          let true_type = match first_type with
              Pitch -> Note
            | _ -> first_type
          in
          (* check that each element in an array literal matches the first type given *)
          let check_call e =
          let (et, e') = expr symbols e in
          let err = "argument of incorrect type included: " ^ string_of_typ first_type ^
            " expected, but " ^ string_of_typ et ^ " found"
          in (check_assign true_type et err, e')
        in
        let args' = List.map check_call l
        in ((Array true_type), SArrayLiteral(args'))
    | CreateArray(t, e) ->
        let (et, e') = expr symbols e in
        let et' =
          (match et with
              Int -> et
            | _ -> raise(Failure("Arrays can only be initialized with an integer length. " ^
                    string_of_expr e ^ " is not an int."))
          )
        in ((Array t), SCreateArray(t, (et', e')))
    | ArrayAccess(s, e) ->
        (* check that brackets are being used with arrays *)
        let (typ, s') = expr symbols s in
        let array_typ = match typ with
            Array t -> t
          | _ ->  raise (Failure ("Only arrays can be accessed with bracket notation. " ^ string_of_expr s ^ " is not an array." ))
        in
        (* check that it is an int inside the bracket access *)
        let (et, e') = expr symbols e in
        let et' = match et with
            Int -> et
          | _ -> raise(Failure("Array indexing must be done with integers. " ^ string_of_expr e ^ " is not an int."))
        in (array_typ, SArrayAccess((typ, s'), (et', e')))

  in

  let check_bool_expr s e =
    let (t', e') = expr s e
    and err = "expected Boolean expression in " ^ string_of_expr e
    in if t' != Bool then raise (Failure err) else (t', e') 
  in

  (* Return a semantically-checked statement i.e. containing sexprs *)
  (* move this function out of the regular check_stmt so that we can maintain our symbol table easier
      especially when working with returns *)
  let rec check_stmt_list symbols typ = function
      [Return _ as s] -> ([fst(check_stmt symbols typ s)], symbols)
    | Return _ :: _   -> raise (Failure "nothing may follow a return")
    | Block sl :: ss  -> check_stmt_list symbols typ (sl @ ss) (* Flatten blocks *)
    | s :: ss         ->
        let (single, new_symbol) = check_stmt symbols typ s in
        let (rest, final_symbols) = check_stmt_list new_symbol typ ss in
        (single :: rest, final_symbols)
    | []              -> ([], symbols)

  and check_stmt symbols typ = function
      Expr e -> (SExpr (expr symbols e), symbols)
    | Return e -> let (t, e') = expr symbols e in
      if t = typ then (SReturn (t, e'), symbols)
      else raise (
  Failure ("return gives " ^ string_of_typ t ^ " expected " ^
     string_of_typ typ ^ " in " ^ string_of_expr e))
    (* A block is correct if each statement is correct and nothing
       follows any Return statement.  Nested blocks are flattened. *)
    | Block sl -> (SBlock(fst (check_stmt_list symbols typ sl)), symbols)
    (* make sure that a variable is neither void nor being rebound *)
    | Bind(lt, id) ->
          if (lt = Void) then
            let err ="cannot bind " ^ id ^ " to Void type"
            in raise (Failure err)
          else
          if (StringMap.mem id symbols) then
            let err = "cannot bind " ^ id ^ " as it already is bound"
            in raise (Failure err)
          else
          let sym' = StringMap.add id lt symbols in ( SBind(lt, id), sym' )
    (* same as above but also checks assignment matches bind *)
    | BindAssign(lt, id, e) as ex ->
          if (lt = Void) then
            let err ="cannot bind " ^ id ^ " to Void type"
            in raise (Failure err)
          else
          if (StringMap.mem id symbols) then
            let err = "cannot bind " ^ id ^ " as it already is bound"
            in raise (Failure err)
          else
            let (rt, e') = expr symbols e in
            let err = "illegal assignment " ^ string_of_typ lt ^ " = " ^
              string_of_typ rt ^ " in " ^ string_of_stmt ex
            in let typ = check_assign lt rt err
            in let sym' = StringMap.add id lt symbols
            in ( SBindAssign(typ, id, (rt, e')), sym' )
    | If(p, b1, b2) -> (SIf(check_bool_expr symbols p, fst(check_stmt symbols typ b1), fst(check_stmt symbols typ b2)), symbols)
    | For(s1, e2, e3, s) ->
      let (sstmt, symbols') = check_stmt symbols typ s1 in
      (SFor( sstmt, check_bool_expr symbols' e2, expr symbols' e3, fst(check_stmt symbols' typ s)), symbols)
    | While(e, s) -> (SWhile(check_bool_expr symbols e, fst(check_stmt symbols typ s)), symbols)

  in

  let check_function func =
    (* Make sure no formals or locals are void or duplicates *)
    check_binds "formal" func.formals;

    (* Build local symbol table of variables for this function *)
    let symbols = List.fold_left (fun m (ty, name) -> StringMap.add name ty m)
                  StringMap.empty (func.formals)
    in
    let (sl, _) = check_stmt symbols func.typ (Block func.body) in

(* body of check_function *)
    { styp = func.typ;
      sfname = func.fname;
      sformals = func.formals;
      sbody = match sl with
          SBlock(sl) -> sl
        | _ -> raise (Failure ("internal error: block didn't become a block?"))
    }

  in (fst (check_stmt_list StringMap.empty Void (List.rev statements)), List.map check_function functions)
